/* **********
 *
 * Boss and Worker classes
 *
 * Copyright (c) 2003 Dan B Goldman, Aaron Hertzmann
 * 
 * This file is part of the BossWorker library.
 *
 * BossWorker is free software; you can redistribute it and/or modify
 * it under the terms of the X11 Software License (see the LICENSE file
 * for details).
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTIBILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the X11
 * License for more details.
 *
 * See BossWorker.h for API documentation
 *
 * **********/

#include <iostream>

#include <unistd.h>
#include <stdio.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>

#include "BossWorker.h"

using namespace std;

#define HireMessage "You're hired!\n"

/*******************************************************************/
/* BossThread declaration                                          */
/*******************************************************************/

class BossThread {
  private:
    pthread_t thread;
    Boss *boss;
    int socket;

    std::string lastMessage;

  public:
    BossThread(pthread_t aThread,
	       Boss *aBoss = NULL,
	       int aSocket = -1)
      : boss(aBoss), socket(aSocket)
      { lastMessage = ""; }

    ~BossThread() { pthread_cancel(thread); }
    int create(const pthread_attr_t &threadAttr);

    static void *handleWorker(void *p); // entry point for receiving thread
    void handleWorker();                // member function for receiving thread

    bool send(const std::string &msg);
};

/********************************************************************/
/* Socket helper code                                               */
/********************************************************************/

int
createListeningSocket(int port)
{
  /* Create and bind a socket that will listen for connections on a port */

    int theSocket = -1;
    int reuse_addr = 1;

    // This socket will allow connections from any ip address to the given port
    struct sockaddr_in address;
    memset((char *) &address, 0, sizeof(address));
    address.sin_family = AF_INET;
    address.sin_port = port;
    address.sin_addr.s_addr = htonl(INADDR_ANY);

    // Set up the socket to receive TCP packets
    theSocket = socket(AF_INET, SOCK_STREAM, 0);
    if (theSocket < 0) {
	perror("socket");
	exit(-1);
    }

    // The same port will be reused for multiple connections
    setsockopt(theSocket, SOL_SOCKET, SO_REUSEADDR, &reuse_addr,
      sizeof(reuse_addr)); 

    // Bind the socket to the address information
    if (bind(theSocket, (struct sockaddr *) &address, sizeof(address)) < 0) {
	perror("bind");
	close(theSocket);
	exit(-1);
    }

    return theSocket;
}

int
createSendingSocket(const string &server, int port)
{
  /* Create and connect a socket that will connect to the given server and port */

    struct hostent *host;
    struct in_addr *addr;
    struct sockaddr_in address;
    int sock, connected;

    // Convert the host string to a host address
    host = gethostbyname(server.c_str());
    if (host != NULL)
	addr = (struct in_addr *) *host->h_addr_list;
    else {
	fprintf(stderr, "Invalid network address \"%s\".\n",server.c_str());
	return -1;
    }

    // This socket is connecting to a port on a specific server
    memset((char *) &address, 0, sizeof(address));
    address.sin_family = AF_INET;
    address.sin_port = (port);
    address.sin_addr.s_addr = addr->s_addr;

    // Create a TCP socket
    sock = socket(AF_INET, SOCK_STREAM, 0);
    
    printf("Connecting to %s on port %d.\n",inet_ntoa(*addr),htons(port));

    // Make the connection
    connected = connect(sock, (struct sockaddr *) &address, sizeof(address));
    if (connected < 0) {
	perror("connect");
	return -1;
    }
    return sock;
}

int
sockWrite(int sockfd, const string &msg)
{
  /* Write a string to a socket.  This is just like the write() system
     call, except that it will make sure all data is transmitted. */

    size_t bytes_sent = 0;
    int this_write;

    const char *buf = msg.data();
    unsigned long count = msg.size();

    // First send the length of message as a long word.
    this_write = write(sockfd, &count, sizeof(count));

    while (bytes_sent < count) {
        // Try to send the rest of the message
	do
	    this_write = write(sockfd, buf, count - bytes_sent);
	while ( (this_write < 0) && (errno == EINTR) );
	if (this_write <= 0)
	    return this_write;  // Socket probably closed
	bytes_sent += this_write;
	buf += this_write;
    }

    return count;
}

int
sockRead(int sockfd, string &msg)
{
  /* Read from a socket until a complete message is received as sent by
     sockWrite.  (A message is preceded by an unsigned long string
     length.)  This function will return -1 if the socket is closed
     during the read operation. */

    int total_count = 0;
    int bytes_read;
    char last_read = 0;

    unsigned long count = 0;

    // Read the message length;
    bytes_read = read(sockfd, &count, sizeof(count));
    if (bytes_read <= 0)
	return -1;
    if (count <= 0) {
	printf("sockRead error: received bad count %d\n",count);
	return -1;
    }
	
    msg.assign("");
    msg.reserve(count);
    while (count--) {
        // Read one character at a time (TODO inefficient)
	bytes_read = read(sockfd, &last_read, 1);
	if (bytes_read <= 0)	    // The other side may have closed unexpectedly
	    return -1;
	msg += last_read;
	total_count++;
    }

    return total_count;
}

void
handleChld(int)
{
  /* This CHLD signal handler waits for all children, so that they don't
     become zombies. */

    int status;
    
    while (wait3(&status, WNOHANG, NULL) > 0);
}

void
installSignalHandlers()
{
  /* Install signal handlers for sockets and pipes */

    struct sigaction chldAction, pipeAction;

    /* Ensure children don't become zombies */
    sigemptyset(&chldAction.sa_mask);
    chldAction.sa_flags = 0;
    chldAction.sa_handler = handleChld;
    sigaction(SIGCHLD, &chldAction, NULL);

    /* Don't kill processes with disconnected pipes */
    sigemptyset(&pipeAction.sa_mask);
    pipeAction.sa_flags = 0;
    pipeAction.sa_handler = SIG_IGN;
    sigaction(SIGPIPE,&pipeAction,NULL);
}

/********************************************************************/
/* Boss implementation                                              */
/********************************************************************/


Boss::Boss(int port)
{
  /* Create a thread to hire workers as they become available */

    totalWorkers = 0;
    totalMessagesSent = totalMessagesReceived = 0;
    totalBytesSent = totalBytesReceived = 0;

    this->port = port;

    pthread_t th;
    size_t size;
    pthread_attr_init(&threadAttr);
    pthread_attr_setstacksize(&threadAttr,PTHREAD_STACK_MIN);
    pthread_attr_getstacksize(&threadAttr,&size);
    printf("stacksize = %d\n",size);

    int result = pthread_create(&th,&threadAttr,hireWorkers,(void *)this);
    pthread_detach(th);

    if (result == 0)
    {
	threads.push_back(new BossThread(th));
	return;
    }

    printf("Error creating boss thread %d\n",result);

}

void *
Boss::hireWorkers (void *p)
{
  /* Wrapper for hiring thread entry point */

    ((Boss *)p)->hireWorkers();

    return NULL;
}

void
Boss::hireWorkers()
{
  /* Hiring thread entry point */

    // Listen for workers on a port
    installSignalHandlers();
    listeningSocket = createListeningSocket(port);
    listen(listeningSocket, 5);

    while (1) {
	int connectedSocket = -1;
	connectedSocket = accept(listeningSocket, NULL, NULL);
	if (connectedSocket < 0) {
	    /* Either a real error occurred, or blocking was interrupted
	       for some reason.  Only abort if a real error occurred. */
	    if (errno != EINTR) {
		perror("accept");
		close(listeningSocket);
		exit(-1);
	    } else
		continue;  /* try this loop again */
	}

	printf("New worker found at connection %d\n", connectedSocket);

	// Spawn a new thread for this worker.
	BossThread *bt = new BossThread(0,this,connectedSocket);
	int result = bt->create(threadAttr);

	if (result == 0) {
	    threads.push_back(bt);
	    totalWorkers++;
	} else
	{
	    printf("Unable to start port handler thread (%d)\n", result);
	    delete bt;
	}
    }

}

Boss::~Boss()
{
  /* Terminate all child threads. */

//    for (ThreadVec::iterator it = threads.begin();
//	 it != threads.end(); ++it)
//         delete *it;
    BossThread *bt;
    while (connections.pop(bt))
    {
	delete bt;
    }

    // close all connections
    close(listeningSocket);

    // TODO does this actually close the other connections?

}

bool
Boss::send(const string &msg)
{
  /* Send message immediately to first available worker */
  /* THIS FUNCTION IS BLOCKING! */

    BossThread *bt;
    connections.pop(bt,true);  // blocks until a worker is available!

    bt->send(msg);

    totalMessagesSent++;
    totalBytesSent += msg.size();

    return true;
}

bool
Boss::receive(string &msg)
{
  /* Receive next available message from worker */
  /* THIS FUNCTION IS BLOCKING! */

  bool val = responses.pop(msg, true);  // blocks until a response is available!
  
  totalMessagesReceived++;
  totalBytesReceived += msg.size();

  return val;
}

int
BossThread::create(const pthread_attr_t &threadAttr)
{
    return pthread_create(&thread, &threadAttr, &handleWorker, this);
}

void *
BossThread::handleWorker(void *p)
{
  /* Wrapper for worker thread entry point */

    ((BossThread *)p)->handleWorker();

    return NULL;
}

void
BossThread::handleWorker()
{
  /* Worker thread entry point */

    // Initialize communications
    string buffer;
    sockWrite(socket,HireMessage);
    printf("welcome message sent to %d\n",socket);

    // Make this worker available for jobs
    boss->connections.push(this);

    // Wait for responses
    while (1) {

	if (sockRead(socket,buffer) < 0) {
	    printf("BOSS SOCKET CLOSING\n");
	    
	    // Resend the corresponding job
	    boss->send(lastMessage);

	    // TODO "this" will become a dangling pointer
	    // should ask the boss to kill us instead.

	    close(socket);
	    pthread_exit(NULL);
	}

#ifndef NDEBUG
	fprintf(stderr,"Boss received %d bytes from worker %d\n",buffer.size(),socket);
#endif
	boss->responses.push(buffer);  // make response available
	boss->connections.push(this);  // make worker available

	fflush(stdout);
    }
}

bool
BossThread::send(const string &msg)
{
  /* Send message to this thread's worker */

    lastMessage = msg;

    bool valid = sockWrite(socket,msg);

#ifndef NDEBUG
    fprintf(stderr,"Boss sent %d bytes to worker %d\n",msg.size(),socket);
#endif

    return valid;
}

string
Boss::status()
{
  char s[1000];

  sprintf(s,"%d workers, read %d/%d, wrote %d/%d",totalWorkers,
	  totalBytesReceived,totalMessagesReceived,
	  totalBytesSent,totalMessagesSent);

  return string(s);
}

/********************************************************************/
/* Worker implementation                                            */
/********************************************************************/

Worker::Worker(const string &bossName, int port)
{
  totalMessagesSent = totalMessagesReceived = 0;
  totalBytesSent = totalBytesReceived = 0;

  /* Initiate a connection to the Boss at the given address and port. */

    printf("Worker opening connection to %s, port %d\n", bossName.c_str(), port);

    socket = createSendingSocket(bossName, port);

    string buffer;
    sockRead(socket,buffer);

    if (buffer != HireMessage)
	printf("WELCOME MESSAGE NOT RECEIVED (got \"%s\" instead)\n",buffer.c_str());
    else
	printf("Welcome message received.\n");

    return;
}

Worker::~Worker()
{
  /* Terminate the connection to the Boss.  It's OK to close the
   socket prematurely, since the boss will resend any jobs previously
   distributed to workers that quit early. */

    close(socket);
}

bool
Worker::receive(string &msg)
{
  /* Get the next job from the Boss */

    int count = sockRead(socket, msg);

    if (count == -1) {
	printf("WORKER SOCKET CLOSED\n");
	fflush(stdout);
	close(socket);
    }

#ifndef NDEBUG
    fprintf(stderr,"Worker received %d bytes from boss.\n",count);
#endif

    totalMessagesReceived++;
    totalBytesReceived += max(0,count);

    return (count > 0);
}

bool
Worker::send(const string &msg)
{
  /* Send the results of a job back to the Boss */

    int count = sockWrite(socket,msg);

    if (count == -1)
	printf("WORKER SOCKET CLOSED\n");

#ifndef NDEBUG
    fprintf(stderr,"Worker sent %d bytes to boss.\n",count);
#endif

    totalMessagesSent++;
    totalBytesSent += max(0,count);

    return (count > 0);
}

string
Worker::status()
{
  char s[1000];

  sprintf(s,"read %d/%d, wrote %d/%d",
	  totalBytesReceived,totalMessagesReceived,
	  totalBytesSent,totalMessagesSent);

  return string(s);
}
