/* **********
 *
 * A thread-safe queue template
 *
 * Copyright (c) 2003 Dan B Goldman, Aaron Hertzmann
 *
 * This file is part of the BossWorker library.
 *
 * BossWorker is free software; you can redistribute it and/or modify
 * it under the terms of the X11 Software License (see the LICENSE file
 * for details).
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTIBILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the X11
 * License for more details.
 *
 * See BossWorker.h for API documentation
 *
 * Dan B Goldman, April 2003
 * dgoldman@cs.washington.edu
 *
 * Thanks to Aaron Hertzmann for the original code
 *
 * **********/

#ifndef _SYNCQUEUE_
#define _SYNCQUEUE_

/* **********

  This header defines a generic thread-safe queue template.  A mutex
  is used to guarantee thread-safety when pushing and popping
  data. This data structure is convenient for safe communications
  across multiple threads.

  Arguments are pass-by-value, not by-reference, so be sure the
  instance type is of a reasonable size.  If large classes are used as
  types, better to store and pass pointers to avoid lots of copying
  and construction.

  By default, the SyncQueue::pop function returns false if the queue
  is empty.  However, it takes an optional boolean argument "blocking"
  which, when true, causes the thread to block until an element is
  pushed into the queue.  Use with care: This can result in deadlock!
 
  ********** */

#include <string>
#include <deque>
#include <pthread.h>

template<class T>
class SyncQueue {
  private:
    std::string name;
    std::deque<T> data;
    pthread_mutex_t mutex;
    pthread_cond_t nonEmpty;
   
  public:
    SyncQueue(const std::string &name = "") {
      /* Create a queue with the given name.
	 (The name is used only for debugging purposes) */

	int result = pthread_mutex_init(&mutex,NULL);
	if (result != 0) {
	    printf("Error initializing mutex!\n");
	    exit(1);
	}
	result = pthread_cond_init(&nonEmpty,NULL);
	if (result != 0) {
	    printf("Error initializing nonEmpty condition variable!\n");
	    exit(1);
	}
	this->name = name;
    }

    ~SyncQueue() {
	pthread_mutex_destroy(&mutex);
	pthread_cond_destroy(&nonEmpty);
    }

    void push(const T &elem) {
      /* Push a new element onto the queue */

	pthread_mutex_lock(&mutex);
	data.push_back(elem);
	// printf("%s QUEUE LENGTH = %d\n",name.c_str(),data.size());
	pthread_cond_signal(&nonEmpty);  // at most one waiting thread awakened
	pthread_mutex_unlock(&mutex);
    }

    bool pop(T &elem, bool blocking=false) {
      /* Pop the next element from the queue.  If blocking is true,
	 the thread will block until an element becomes available. */

	bool nonempty;

	pthread_mutex_lock(&mutex);
	
	if (blocking)
	    while (!data.size())
		pthread_cond_wait(&nonEmpty,&mutex);

	if (data.size()) {
	    elem = data.front();
	    data.pop_front();
	    nonempty = true;
	} else
	    nonempty = false;

	pthread_mutex_unlock(&mutex);

	return nonempty;
    }
};

#endif // _SYNCQUEUE_
